#!/usr/bin/env -S python #
# -*- coding: utf-8 -*-

"""
Provides the `Keyhole` class and attendant exception classes,
for simple line-by-line serial-port communication (especially
with microcontrollers that have been programmed using the
Keyhole library for the Arduino IDE).
"""

__version__ = '1.4.2'

__all__ = [
	'Keyhole',
	'KeyholeError', 'KeyholeVariableError', 'KeyholeValueError', 
	'KEOF',
]

from ast import literal_eval
try:    from time import perf_counter as clock
except: from time import clock

import serial

try: from BCI2000Tools.Container import Bunch as Container
except: Container = dict

class KeyholeError( Exception ): pass
class KeyholeVariableError( AttributeError, KeyholeError ): pass
class KeyholeValueError( ValueError, KeyholeError ): pass
class KEOF: pass
KEOF = KEOF()

class Keyhole( object ):
	"""
	Connect to a serial device using the same constructor arguments
	as `serial.Serial`::
	
	    k = Keyhole('COM4')
	
	Send a message using the `<<` operator. It will be automatically
	encoded to `bytes` and a newline will be appended if there wasn't
	one already::
	
	    k << "hello"     # sends b'hello\n'
	
	The return value will provide information from the first line of
	any reply that occurs within the timeout period (decoded into a
	dict, if the line was a valid dict literal) along with timing
	information.
	
	If the microcontroller has been programmed using the Keyhole
	library for the Arduino IDE, and it exposes variables via the
	keyhole, then these variables can be accessed as attributes::
	
	    print(k.foo)
	    k.foo = 123
	
	To close the connection, either `k<<KEOF` or `k._close()` will
	work (this will also happen automatically when the `k` is
	garbage-collected).	
	"""
	def __init__( self, port, timeout=0.1, encoding='UTF8', **kwargs ):
		"""
		Use the same constructor args as `serial.SerialBase`.
		"""
		self.__dict__.update(
			_connection = serial.Serial( port, timeout=float( timeout ), **kwargs ),
			_encoding  = encoding,
			_lastReply  = None,
		)
		
	def _close( self ):
		"""You can also do `self<<KEOF`."""
		return self._connection.close()
	
	def __getattr__( self, name ):
		reply = self << name
		if name in reply: return reply[ name ]
		if '_KEYHOLE_ERROR_MSG' in reply or 'KEYHOLE_COMMAND_ERROR' in reply: # the latter is from an older version of the Arduino Keyhole library
			raise KeyholeVariableError( 'failed to recognize %r' % name )
		raise KeyholeError( 'could not understand reply' )
		
	def __setattr__( self, name, value ):
		reply = self << name + '=' + repr( value )
		if '_KEYHOLE_ERROR_MSG' in reply:
			errorType = reply.get( '_KEYHOLE_TYPE', None )
			if errorType in [ 'ReadOnly' ]: raise KeyholeVariableError( reply[ '_KEYHOLE_ERROR_MSG' ] )
			if errorType in [ 'BadValue' ]: raise KeyholeValueError(    reply[ '_KEYHOLE_ERROR_MSG' ] )
			raise KeyholeVariableError( 'failed to set %r' % name )
		elif 'KEYHOLE_COMMAND_ERROR' in reply: # from an older version of the Arduino Keyhole library
			raise KeyholeVariableError( 'failed to set %r' % name )
	
	def _update( self, **kwargs ):
		for k, v in kwargs.items(): setattr( self, k, v )
			
	def __dir__( self ):
		reply = self << '?'
		return [ k for k in reply.keys() if not k.startswith( '_' ) ]
	_getAttrNames = __dir__
	
	def __lshift__( self, command ):
		if command is KEOF: self._connection.close(); return Container( _CLOSED=True )
		command = ( command.rstrip( '\n' ) + '\n' ).encode( self._encoding )
		write = self._connection.write
		flush = self._connection.flushOutput
		read  = self._connection.readline		
		
		########
		t0 = clock()
		write( command )
		flush()
		reply = read()
		elapsedMilliseconds = ( clock() - t0 ) * 1000
		########
		
		reply = reply.decode( self._encoding )
		if reply.strip().startswith( '{' ):
			try:    reply = Container( literal_eval( reply ) )
			except: reply = Container( _REPLY=reply )
		else:
			reply = Container( _REPLY=reply )
		reply[ '_ELAPSED_MS' ] = elapsedMilliseconds
		self.__dict__.update( _lastReply=reply )
		return reply
		
if __name__ == '__main__':
	import sys
	args = sys.argv[ 1: ]
	if not args: args.append( 'COM7' )
	k = Keyhole( *args )
	print( k << '?' )
	